from .Datetime import Datetime
from .dtrange import dtrange
from .utils import date_from_datetime

class drange(object):
    '''
    Date iterator.
    '''

    def __init__(self, start, stop=None, step=None, n=None, units='d',
                 endpoint=False, calendar='gregorian'):
        '''
        Parameters
        ----------
        start : date
            First inclusive datetime.
        stop : date, optional
            Optional. Inclusive if closed=True.
        step : timedelta or number, optional
            Requires units if is a number.
        n : number, optional
            Number of divisions between start and stop.
        units : str, optional
            y=year, m=month, w=week, d=day
            Default is 'd'.
        endpoint : bool, optional
            If true, then stop is included, otherwise use half-open bounds.
            Default is False.
        calendar : str
            One of [360, gregorian, julian, leap, noleap].
            Default is gregorian.
        '''
        self.dtrange = dtrange(start, stop=stop, step=step, n=n, units=units,
                               endpoint=endpoint, calendar=calendar)

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        self.dtrange.next()
        if 'gregorian' != self.dtrange.calendar:
            return Datetime(self.dtrange.current.year,
                            self.dtrange.current.month,
                            self.dtrange.current.day)
        else:
            return date_from_datetime(self.dtrange.current)


#     LICENSE BEGIN
#
#     dtrange - Datetime, date and time range iterators.
#     Copyright (C) 2016  Remik Ziemlinski
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#     LICENSE END



from .calendar import days_in_month, days_in_year, date_plus_days
from datetime import datetime, timedelta
from .utils import timedelta_from_uniform_units


class dtrange(object):
    '''
    datetime iterator.
    '''

    def __init__(self, start, stop=None, step=None, n=None, units='d',
                 endpoint=False, calendar='gregorian'):
        '''
        Parameters
        ----------
        start : datetime
            First inclusive datetime.
        stop : datetime, optional
            Optional. Inclusive if closed=True.
        step : timedelta or number, optional
            Requires units if is a number.
        n : number, optional
            Number of divisions between start and stop.
        units : str, optional
            y=year, m=month, w=week, d=day, h=hour, min=minutes
            s=seconds, ms=milliseconds, us=microseconds
            Default is 'd'.
        endpoint : bool, optional
            If true, then stop is included, otherwise use half-open bounds.
            Default is False.
        calendar : str
            One of [360, gregorian, julian, leap, noleap].
            Default is gregorian.
        '''
        self.calendar = calendar
        self.current = start
        self.endpoint = endpoint
        self.first_iteration = True
        self.n = n
        self.non_uniform_units = None
        self.step = step
        self.stop = stop

        def parse_step_as_number():
            if type(step) != timedelta:
                if not units:
                    raise Exception('Expected units.')
                if units in 'my':
                    self.non_uniform_units = units
                else:
                    self.step = timedelta_from_uniform_units(units)

        if stop:
            if step:
                parse_step_as_number()
            elif n:
                if endpoint:
                    n -= 1
                self.step = (stop - start) / n
            else:
                raise Exception('Expected step or n.')
        elif step and units and n:
            parse_step_as_number()
        else:
            raise Exception('Expected (stop,step|n) or (step,units,n).')

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if not self.first_iteration:
            if self.non_uniform_units:
                self.next_non_uniform()
            else:
                self.next_uniform()
        else:
            self.first_iteration = False

        return self.next_return()

    def next_non_uniform(self):
        if 'y' == self.non_uniform_units:
            self.next_non_uniform_year()
        else:
            self.next_non_uniform_month()

    def next_non_uniform_month(self):
        for _ in range(self.step):
            days = days_in_month(self.current.month,
                                 self.current.year,
                                 self.calendar)
            self.current = date_plus_days(self.current, days, self.calendar)

    def next_non_uniform_year(self):
        for _ in range(self.step):
            days = days_in_year(self.current.year, self.calendar)
            self.current = date_plus_days(self.current, days, self.calendar)

    def next_return(self):
        if self.endpoint:
            if self.current > self.stop:
                raise StopIteration
        else:
            if self.stop:
                if self.current >= self.stop:
                    raise StopIteration
            elif self.n <= 0:
                raise StopIteration
            else:
                self.n -= 1

        return self.current

    def next_uniform(self):
        if 0 == self.step.microseconds + self.step.seconds:
            self.current = date_plus_days(self.current, self.step.days,
                                          self.calendar)
        else:
            # TODO: Exotic calendar time arithmetic.
            self.current += self.step


#     LICENSE BEGIN
#
#     dtrange - Datetime, date and time range iterators.
#     Copyright (C) 2016  Remik Ziemlinski
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#     LICENSE END



from .calendar import ordinal


def time_ordinal(dt):
    '''Return fraction of day [0-1].'''
    if hasattr(dt, 'hour'):
        t = dt.hour*3600. + dt.minute*60. + dt.second + dt.microsecond/1.0e6
        return t / 86400.
    else:
        return 0.

def dtfraction(lower, mid, upper, calendar='gregorian'):
    '''Return fraction [0,1] for datetime between lower and upper bounds.'''
    ld = ordinal(lower, calendar) + time_ordinal(lower)
    md = ordinal(mid,   calendar) + time_ordinal(mid)
    ud = ordinal(upper, calendar) + time_ordinal(upper)
    denom = ud - ld
    if denom:
        return (md - ld) / denom
    else:
        return 0.

def dfraction(lower, mid, upper, calendar='gregorian'):
    return dtfraction(lower, mid, upper, calendar)

def tfraction(lower, mid, upper):
    lt = time_ordinal(lower)
    mt = time_ordinal(mid)
    ut = time_ordinal(upper)
    denom = ut - lt
    if denom:
        return (mt - lt) / denom
    else:
        return 0.



def find_bounds(items, item):
    ''' Binary search bracketing indices.

    Parameters
    ----------
    items : list
        Ascending sorted items.
    item : comparable object

    Returns
    -------
    [lower, upper) half-open indices
    '''
    n_1 = len(items) - 1
    if n_1 < 0:
        return None, None
    if item <= items[0]:
        return 0,0
    if item >= items[-1]:
        return n_1, n_1

    left, right = 0, n_1
    while left <= right:
        mid = int((left + right) / 2)
        if items[mid] == item:
            break
        elif items[mid] < item:
            left = mid + 1
        else:
            right = mid - 1

    if right < left:
        left = right
    else:
        left = mid

    right = min(n_1, left + 1)
    if items[left] <= item and item <= items[right]:
        return left, right
    else:
        return None, None



from .Datetime import Datetime
from datetime import date, datetime, timedelta

DAYS_IN_MONTH = (0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
DAYS_IN_MONTH_LEAP = (0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31)
DAYS_BEFORE_MONTH = (0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334)
DAYS_BEFORE_MONTH_LEAP = (0, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335)

def day_of_year(dt, calendar):
    '''
    >>> day_of_year(datetime(2000, 2, 1), 'julian')
    32
    '''
    dim = days_in_months(dt.year, calendar)
    result = dt.day
    m = 1
    while m < dt.month:
        result += dim[m]
        m += 1

    return result

def day_of_year_date(doy, year, calendar):
    '''
    >>> day_of_year_date(32, 2000, 'julian')
    2000, 2, 1
    >>> day_of_year_date(367, 2000, 'julian')
    2001, 1, 1
    '''
    done = False
    while 1:
        dim = days_in_months(year, calendar)
        for month in range(1, 13):
            if doy <= dim[month]:
                done = True
                break
            doy -= dim[month]
        if done:
            break
        year += 1

    return year, month, doy

def date_plus_days(dt, days, calendar):
    '''Add days (integer) to a date for calendar.'''
    if 'gregorian' == calendar:
        return dt + timedelta(days=days)

    doy = day_of_year(dt, calendar)
    future = doy + days
    y,m,d = day_of_year_date(future, dt.year, calendar)

    if type(dt) == datetime:
        result = datetime(y, m, d, dt.hour, dt.minute, dt.second, dt.microsecond, dt.tzinfo)
    elif 'gregorian' != calendar:
        # Python 'date' fails when 30 days are used in Feb.
        result = Datetime(y, m, d)
    else:
        result = date(y, m, d)

    return result

def days_in_month(m, y, calendar):
    if 'gregorian' == calendar:
        if is_leap_year_gregorian(y):
            return DAYS_IN_MONTH_LEAP[m]
        else:
            return DAYS_IN_MONTH[m]
    elif 'julian' == calendar:
        if is_leap_year_julian(y):
            return DAYS_IN_MONTH_LEAP[m]
        else:
            return DAYS_IN_MONTH[m]
    elif 'leap' == calendar:
        return DAYS_IN_MONTH_LEAP[m]
    elif 'noleap' == calendar:
        return DAYS_IN_MONTH[m]
    elif '360' == calendar:
        return 30
    else:
        raise Exception('Unsupported calendar={0}'.format(calendar))

def days_in_months(y, calendar):
    '''Return array of days per month (index 1 is January).'''
    if is_leap_year(y, calendar):
        return DAYS_IN_MONTH_LEAP
    elif '360' == calendar:
        return [30]*13
    else:
        return DAYS_IN_MONTH

def days_in_year(y, calendar):
    if 'gregorian' == calendar:
        if is_leap_year_gregorian(y):
            return 366
        else:
            return 365
    elif 'julian' == calendar:
        if is_leap_year_julian(y):
            return 366
        else:
            return 365
    elif 'leap' == calendar:
        return 366
    elif 'noleap' == calendar:
        return 365
    elif '360' == calendar:
        return 360
    else:
        raise Exception('Unsupported calendar={0}'.format(calendar))

def is_leap_year(y, calendar):
    if 'gregorian' == calendar:
        return is_leap_year_gregorian(y)
    elif 'julian' == calendar:
        return is_leap_year_julian(y)
    elif 'leap' == calendar:
        return True
    else:
        return False

def is_leap_year_gregorian(y):
    if y % 400 == 0:
        return True
    elif y % 100 == 0:
        return False
    elif y % 4 == 0:
        return True
    else:
        return False

def is_leap_year_julian(y):
    return y % 4 == 0

def leap_years_before(y, calendar):
    '''Return number of years before input year that are leap.'''
    y -= 1
    if 'gregorian' == calendar:
        return y * 0.2425
    elif 'julian' == calendar:
        return y/4
    elif 'leap' == calendar:
        return y
    else:
        return 0

def ordinal(dt, calendar):
    '''
    >>> ordinal(datetime(1,1,1), 'gregorian')
    1
    >>> ordinal(datetime(2,1,1), 'gregorian')
    366
    '''
    if 'gregorian' == calendar:
        t = type(dt)
        if t not in (date, datetime):
            dt = datetime(dt.year, dt.month, dt.day)
        return dt.toordinal()
    elif 'julian' == calendar:
        if is_leap_year_julian(dt.year):
            mdays = DAYS_BEFORE_MONTH_LEAP[dt.month]
        else:
            mdays = DAYS_BEFORE_MONTH[dt.month]
        nleap = int(dt.year/4)
        return 366*nleap + 365*(dt.year-nleap-1) + mdays + dt.day
    elif 'noleap' == calendar:
        return 365*(dt.year-1) + DAYS_BEFORE_MONTH[dt.month] + dt.day
    elif 'leap' == calendar:
        return 366*(dt.year-1) + DAYS_BEFORE_MONTH_LEAP[dt.month] + dt.day
    else:
        return 360*(dt.year-1) + 30*(dt.month-1) + dt.day




from datetime import date, timedelta


def date_from_datetime(dt):
    return date(dt.year, dt.month, dt.day)

def timedelta_from_uniform_units(units):
    '''Return timedelta for units [day, week, sec, hour, min, us, ms].'''
    c = units[0]
    if 'd' == c:
        r = timedelta(1)
    elif 'w' == c:
        r = timedelta(7)
    elif 's' == c:
        r = timedelta(seconds=1)
    elif 'h' == c:
        r = timedelta(hours=1)
    elif 'us' == units:
        r = timedelta(microseconds=1)
    elif 'ms' == units:
        r = timedelta(milliseconds=1)
    elif 'min' == units:
        r = timedelta(minutes=1)
    else:
        raise Exception('Unsupported timedelta units={0}'.format(units))

    return r




from . import calendar
from . import utils
from .bounds import find_bounds
from .Datetime import Datetime
from .drange import drange
from .dtrange import dtrange
from .fraction import dtfraction, dfraction, tfraction
from .trange import trange

__version__ = '1.2.2'



class Datetime(object):
    '''A container that won't fail like datetime with exotic calendars.'''
    def __init__(self, y, m, d, h=0, mn=0, s=0, us=0):
        self.year = y
        self.month = m
        self.day = d
        self.hour = h
        self.minute = mn
        self.second = s
        self.microsecond = us

    def __cmp__(self, other):
        return self.cmp(other)

    def __eq__(self, other):
        return self.cmp(other) == 0

    def __ge__(self, other):
        return self.cmp(other) >= 0

    def __gt__(self, other):
        return self.cmp(other) > 0

    def __le__(self, other):
        return self.cmp(other) <= 0

    def __lt__(self, other):
        return self.cmp(other) < 0

    def __ne__(self, other):
        return self.cmp(other) != 0

    def __repr__(self):
        return 'dtrange.Datetime(%d, %d, %d, %d, %d, %d, %d)' % \
                (self.year, self.month, self.day,
                 self.hour, self.minute, self.second, self.microsecond)

    def __str__(self):
        second = self.second
        if self.microsecond:
            subsec = self.microsecond / 1.0e6
            second += int(subsec) # In case overflow.
            frac = subsec % 1
            micro = ('%f' % (frac)).lstrip('0')
        else:
            micro = ''

        return '%04d-%02d-%02d %02d:%02d:%02d%s' % \
                (self.year, self.month, self.day,
                 self.hour, self.minute, second, micro)

    def cmp(self, other):
        if hasattr(other, 'second'):
            return self.cmp_datetime(other.year, other.month, other.day,
                                     other.hour, other.minute, other.second,
                                     other.microsecond)
        else:
            return self.cmp_date(other.year, other.month, other.day)

    def cmp_date(self, y, m, d):
        c = self.year - y
        if c: return c

        c = self.month - m
        if c: return c

        c = self.day - d
        if c: return c

        return 0

    def cmp_datetime(self, y, m, d, h, mn, s, us):
        c = self.cmp_date(y, m, d)
        if c: return c

        c = self.hour - h
        if c: return c

        c = self.minute - mn
        if c: return c

        c = self.second - s
        if c: return c

        c = self.microsecond - us
        if c: return c

        return 0




import os
from setuptools import setup, find_packages

try:
    from pypandoc import convert
    readmd = lambda f: convert(f, 'rst')
except:
    readmd = lambda f: open(f).read()

# Allow setup.py to be run from any path.
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    author='Remik Ziemlinski',
    author_email='first.last@gmail.com',
    description='Multi-calendar datetime, date, and time range iterators.',
    download_url='https://github.com/rsmz/dtrange/archive/dtrange-1.2.2.tar.gz',
    license='GPLv3',
    long_description=readmd('README.md'),
    name='dtrange',
    packages=find_packages(exclude=['*test*']),
    url='https://www.github.com/rsmz/dtrange',
    version='1.2.2',
)
